-- This script is used to automatically edit SQL Server Agent job steps so
-- unescaped tokens are prefaced with the ESCAPE_NONE syntax that was added in
-- SQL Server 2005 SP1.

--if (@@microsoftversion < 0x90007FE)
--BEGIN
--    RAISERROR('You should only run this script on at least SQL Server 2005 SP1.', 20, 127) WITH LOG
--    return
--END

use msdb
go

if exists (select * from sys.objects where name = N'fn_PrefaceTokensWithEscapeNone' and type = 'FN')
    drop function fn_PrefaceTokensWithEscapeNone
go

-- This function changes @commands so all bare tokens
-- are prefaced with the ESCAPE_NONE syntax.
create function fn_PrefaceTokensWithEscapeNone(@commands nvarchar(max)) RETURNS nvarchar(max)
AS
BEGIN
    declare @idxTokenStart int
    select @idxTokenStart = CHARINDEX(N'$(', @commands)
    while (@idxTokenStart != 0 and @idxTokenStart is not null)
    BEGIN
        declare @idxCloseParen int
        select @idxCloseParen = CHARINDEX(N')', SUBSTRING(@commands, @idxTokenStart, LEN(@commands)))
        -- Error checking. If no closing parenthesis exists, return.
        if (0 = @idxCloseParen)
        BEGIN
            return @commands
        END

        -- Deduce the token variable.
        declare @tokenLen int
        select @tokenLen = @idxCloseParen - LEN(N'$(') - 1
        declare @token nvarchar(max)
        select @token = SUBSTRING(@commands, @idxTokenStart + LEN(N'$('), @tokenLen)

        -- Determine whether @token contains a mismatched number of opening and
        -- closing parenthesis. This behavior could occur if invalid syntax is
        -- in a comment block. If the syntax is in a comment block, go to the next token.
        declare @idx int
        declare @cOpenParens int
        declare @cCloseParens int

        select @cOpenParens = 0
        select @idx = CHARINDEX(N'(', @token);
        while (@idx != 0)
        BEGIN
            select @cOpenParens = @cOpenParens + 1
            select @idx = CHARINDEX(N'(', @token, @idx + 1);
        END

        select @cCloseParens = 0
        select @idx = CHARINDEX(N')', @token);
        while (@idx != 0)
        BEGIN
            select @cCloseParens = @cCloseParens + 1
            select @idx = CHARINDEX(N')', @token, @idx + 1);
        END

        -- Special case for the WMI token.
        if (N'WMI(' = SUBSTRING(@token, 1, LEN(N'WMI(')))
        BEGIN
                select @cOpenParens = @cOpenParens - 1
        END

        if ((@cOpenParens = @cCloseParens) and
            (N'ESCAPE_NONE(' != SUBSTRING(@token, 1, LEN(N'ESCAPE_NONE('))) and
            (N'ESCAPE_SQUOTE(' != SUBSTRING(@token, 1, LEN(N'ESCAPE_SQUOTE('))) and
            (N'ESCAPE_DQUOTE(' != SUBSTRING(@token, 1, LEN(N'ESCAPE_DQUOTE('))) and
            (N'ESCAPE_RBRACKET(' != SUBSTRING(@token, 1, LEN(N'ESCAPE_RBRACKET('))))
        BEGIN
            select @commands = STUFF(@commands, @idxTokenStart + LEN(N'$('), @tokenLen, N'ESCAPE_NONE(' + @token + N')')
        END
        select @idxTokenStart = CHARINDEX(N'$(', @commands, @idxTokenStart + 1)
    END

    return @commands

END
go

if exists (select * from sys.objects where name = N'sp_AddEscapeNoneToJobStepTokens' and type = 'P')
    drop procedure sp_AddEscapeNoneToJobStepTokens
go

-- This procedure lets you update jobs so bare tokens
-- are prefaced with the ESCAPE_NONE syntax. By default, all jobs are updated.
-- You can optionally specify @job_name, @job_id, or @owner_name
-- to limit the jobs that are affected.
CREATE PROCEDURE sp_AddEscapeNoneToJobStepTokens
(
    @job_name nvarchar(128) = null,
    @job_id uniqueidentifier = null,
    @owner_name nvarchar(256) = null
)
AS


-- Find the jobs to update. These jobs must match all the input
-- criteria, unless all the inputs are null. In that case,
-- examine all jobs. The jobs must also be jobs that are created locally,
-- such as sysjobs.originating_server_id = 0. The job cannot be a job that we run
-- because another server told us to run the job. Additionally, if the job
-- is local but it was meant to be run on a target server, we send an
-- update for the job.
declare @jobsToUpdate TABLE (job_id uniqueidentifier not null)

insert into @jobsToUpdate
    select job_id
    from sysjobs
    where originating_server_id = 0 -- local jobs
    and ((COALESCE(@job_name, sysjobs.name) = sysjobs.name) and
         (COALESCE(@job_id, sysjobs.job_id) = sysjobs.job_id) and
         (COALESCE(@owner_name, suser_sname(sysjobs.owner_sid)) = suser_sname(sysjobs.owner_sid)))

-- Find the job steps to update, and create the new command by using
-- the fn_PrefaceTokensWithEscapeNone function.
declare @jobStepsToUpdate TABLE (job_id uniqueidentifier not null,
                                 step_id int not null,
                                 command_old nvarchar(max) null,
                                 command_new nvarchar(max) null)

insert into @jobStepsToUpdate
(job_id, step_id, command_old, command_new)
    select job_id, step_id, command, dbo.fn_PrefaceTokensWithEscapeNone(command)
    from sysjobsteps
    where sysjobsteps.job_id =
    (select job_id
     from @jobsToUpdate
     where job_id = sysjobsteps.job_id)

-- Update the actual job step commands. This occurs first before
-- the updated jobs are pushed out to the target servers so the
-- target servers actually receive the updated version.
declare @updated_job_id uniqueidentifier
declare @updated_job_step_id int
declare @updated_job_step_command nvarchar(max)

declare job_steps_cursor CURSOR FOR
    select job_id, step_id, command_new
    from @jobStepsToUpdate
    order by job_id, step_id

OPEN job_steps_cursor
FETCH NEXT from job_steps_cursor into @updated_job_id, @updated_job_step_id, @updated_job_step_command
WHILE (@@FETCH_STATUS <> -1)
BEGIN
    IF (@@FETCH_STATUS <> -2)
    BEGIN
        EXEC sp_update_jobstep @job_id = @updated_job_id, @step_id = @updated_job_step_id, @command = @updated_job_step_command
    END
    FETCH NEXT from job_steps_cursor into @updated_job_id, @updated_job_step_id, @updated_job_step_command
END

CLOSE job_steps_cursor
DEALLOCATE job_steps_cursor


-- For multiserver jobs, call the sp_post_msx_operation stored procedure to update
-- all the target servers. Note: The sp_post_msx_operation stored procedure is safe
-- to call because it checks if the job is really a multiserver job or not.
declare jobs_cursor CURSOR FOR
    select job_id
    from @jobsToUpdate

open jobs_cursor
FETCH NEXT from jobs_cursor into @updated_job_id
WHILE (@@FETCH_STATUS <> -1)
BEGIN
    IF (@@FETCH_STATUS <> -2)
    BEGIN
        EXEC sp_post_msx_operation @operation = 'UPDATE', @job_id = @updated_job_id
    END
    FETCH NEXT from jobs_cursor into @updated_job_id
END

CLOSE jobs_cursor
DEALLOCATE jobs_cursor

select N'Warning' = N'The following job steps were analyzed and potentially updated to add the ESCAPE_NONE macro before any job tokens that were not already escaped. Review the modified job steps, and replace ESCAPE_NONE with the correct escape macro.'

-- List the jobs that were processed, and include the previous command text.
select suser_sname(jobs.owner_sid) as N'Job owner', jobs.name as N'Job name', jobs.job_id, jobStepsUpdated.step_id, N'Modified' = CASE WHEN jobStepsUpdated.command_new != jobStepsUpdated.command_old THEN 1 ELSE 0 END, N'Command' = jobStepsUpdated.command_new, N'Previous Command' = jobStepsUpdated.command_old
    from sysjobs as jobs, @jobsToUpdate as jobsUpdated, @jobStepsToUpdate as jobStepsUpdated
    where jobsUpdated.job_id = jobs.job_id and jobsUpdated.job_id = jobStepsUpdated.job_id
    order by 'Job name', jobStepsUpdated.step_id
go
